#!/usr/bin/env python3

import sys
import os
import argparse
import logging
from multiprocessing import Process
from time import sleep

import numpy as np

from arts_tools.constants import NTAB, NSB, SB_TABLE

# check psrdada import as psrdada-python is not an explicit requirement in setup.py
try:
    import psrdada
except ImportError:
    raise ImportError("Cannot import psrdada; Is psrdada-python installed?")

#: Number of polarizations
NPOL = 4
#: Number of channels
NCHAN = 1536
#: Number of samples per PSRDADA page
NSAMP = 12500


def get_psrdada_header(fname):
    """
    Read header raw PSRDADA file

    :param str fname: Path to PSRDADA file
    :return: header (dict)
    """
    # load a typical amount of bytes from the file and look for header size keyword
    nbyte = 1
    raw_header = ''
    with open(fname, 'r') as f:
        while True:
            raw_header = raw_header + f.read(nbyte)
            header = [line.strip().split(maxsplit=1) for line in raw_header.split('\n')]
            header = np.array(header)
            try:
                key_index = np.where(header == 'HDR_SIZE')[0]
                hdr_size = header[key_index, 1][0].astype(int)
            except (IndexError, ValueError):
                if nbyte > 1e6:
                    logging.error(f"Key HDR_SIZE not found in first MB of PSRDADA file {fname}")
                    sys.exit(1)
                nbyte += 4096
            else:
                break
        # load the full header with known size
        f.seek(0)
        header = f.read(hdr_size)
        # convert to dict, skipping empty lines and zero padding at the end
        header = dict([line.strip().split(maxsplit=1) for line in header.split('\n') if line][:-1])
    return header


def read_psrdada_to_buffer(files, header, key, dadafits):
    """
    Read PSRDADA files into a memory buffer

    :param list files: Files to read into memory buffer
    :param dict header: PSRDADA header
    :param str key: PSRDADA shared memory key
    :param Process dadafits: Handle to dadafits process
    """
    # convert key to hexadecimal
    hexkey = int(key, 16)

    # connect to the memory buffer
    writer = psrdada.Writer()
    writer.connect(hexkey)

    # write header
    writer.setHeader(header)

    # sort input files
    files.sort()

    # loop over files and read them into header
    for i, fname in enumerate(files):
        with open(fname, 'rb') as f:
            # skip header
            f.seek(int(header['HDR_SIZE']))
            # read data page by page
            while True:
                # check if dadafits is still alive to read the data we are about to write
                if not dadafits.is_alive():
                    # not calling writer.disconnect here, as it hangs forever. Unclear why
                    # raise an error here, as it is caught and logged by the main programme
                    raise ChildProcessError("dadafits no longer running; not writing new data")

                data = np.fromfile(f, count=int(header['RESOLUTION']), dtype='uint8')
                # if at end of file, break loop
                if len(data) == 0:
                    break
                # write page into buffer
                page = np.asarray(writer.getNextPage())
                page[:] = data[:]
                # mark the buffer as filled
                writer.markFilled()

    # done, mark end of data, then disconnect
    writer.getNextPage()
    writer.markEndOfData()
    writer.disconnect()


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description=f"Convert raw IQUV data to PSRFITS.\n"
                                                 f"The output can either be one file for each "
                                                 f"of the {NTAB} tied-array beams (TABs),\nor one "
                                                 f"file for each user-specified synthesized beam (SB) "
                                                 f"between 0 and {NSB}.")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--tab', action='store_true', help="Create TABs instead of SBs")
    group.add_argument('--sb', type=int, nargs='*', help="Space-separated list of SBs to create")

    parser.add_argument('--output_dir', required=True, help="Output directory")
    parser.add_argument('--key', default='dada', help="4-digit hexadecimal Key to use for PSRDADA shared "
                                                      "memory buffer (Default: %(default)s)")
    parser.add_argument('--templates', help="Path to folder with FITS templates "
                                            "(Default: use templates provided with package)")
    parser.add_argument('--sbtable', help="Path to SB table "
                                          "(Default: use SB table provided with package)")
    parser.add_argument('--verbose', '-v', action='store_true', help="Increase verbosity")

    parser.add_argument('raw_files', nargs='+', help="Path to input PSRDADA file(s) belonging to one observation. "
                                                     "If multiple files are given, they are sorted by filename "
                                                     "before processing")

    # print help if no arguments are given
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    # set log verbosity
    if args.verbose:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO
    logging.basicConfig(format='%(levelname)s: %(message)s', level=logging_level, stream=sys.stderr)

    logging.debug(f"Arguments: {args}")

    # verify the input files exist
    for fname in args.raw_files:
        if not os.path.isfile(fname):
            logging.error(f"Input file does not exist: {fname}")
            sys.exit(1)

    # set FITS templates dir and SB table dir
    root_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
    if args.templates is None:
        args.templates = os.path.join(root_dir, 'static', 'fits_templates')
        logging.debug(f"Setting FITS templates directory to {args.templates}")
    if args.sbtable is None:
        args.sbtable = os.path.join(root_dir, SB_TABLE)
        logging.debug(f"Setting SB table to {args.sbtable}")

    # set SB command
    if args.tab:
        sb_cmd = ""
    else:
        # check if SB range is valid
        if (np.any(args.sb) > NSB) or np.any(args.sb) < 0:
            logging.error(f"Invalid SB index given, allowed values are 0 to {NSB}")
            sys.exit(1)
        sb_cmd = f"-S {args.sbtable} -s " + ','.join(map(str, args.sb))

    # create output directory
    if os.path.isdir(args.output_dir):
        logging.warning(f"Output directory already exists: {args.output_dir}")
    try:
        os.makedirs(args.output_dir, exist_ok=True)
    except Exception as e:
        logging.error(f"Failed to create output directory: {e}")
        sys.exit(1)

    # get PSRDADA header from first file
    header = get_psrdada_header(args.raw_files[0])

    # create memory buffer (single page, as processing is slower than reading data anyway)
    buffer_size = NTAB * NCHAN * NSAMP * NPOL
    cmd = f"dada_db -a {header['HDR_SIZE']} -b {buffer_size} -n 1 -k {args.key} -w 2>/dev/null"
    logging.debug(f"dada_db command: {cmd}")
    memory_buffer = Process(target=os.system, args=(cmd, ), name='dada_db')
    logging.info(f"Creating PSRDADA memory buffer of size {buffer_size / 1e9} GB")
    memory_buffer.start()
    sleep(1)

    # if the memory buffer is not alive, something failed (probably key that was already in use)
    if not memory_buffer.is_alive():
        logging.error(f"Failed to set up memory buffer, perhaps PSRDADA key is already in use? "
                      f"Run \"dada_db -d -k {args.key}\" to remove it (But be careful not to remove "
                      f"buffers used by observations when running this script on any of arts001 - arts040)")
        sys.exit(1)

    # run FITS writer
    cmd = f"dadafits -k {args.key} -l /dev/null -t {args.templates} -d {args.output_dir} {sb_cmd}"
    logging.debug(f"dadafits command: {cmd}")
    dadafits = Process(target=os.system, args=(cmd, ), name='dadafits')
    logging.info(f"Starting dadafits")
    dadafits.start()
    sleep(1)

    # read data into buffer
    try:
        read_psrdada_to_buffer(args.raw_files, header, args.key, dadafits)
    except Exception as e:
        logging.error(f"Exception in writing data to memory buffer: {type(e).__name__}: {e}")
        # remove the running Processes
        dadafits.terminate()
        # sometimes terminating the buffer Process does not actually remove the buffer, so
        # remove it explicitly first
        cmd = f"dada_db -d -k {args.key} >/dev/null 2>&1"
        os.system(cmd)
        memory_buffer.terminate()
        sys.exit(1)

    # when done reading data into buffer, wait until dadafits exits
    logging.info("Waiting for dadafits to finish")
    dadafits.join()
    logging.info("dadafits finished")

    # remove the memory buffer
    logging.info("Removing PSRDADA memory buffer")
    # sometimes terminating the buffer Process does not actually remove the buffer, so
    # remove it explicitly first
    cmd = f"dada_db -d -k {args.key} >/dev/null 2>&1"
    os.system(cmd)
    memory_buffer.terminate()

    logging.info("Done")



#!/usr/bin/env python3

import os
import urllib
import ssl

import numpy as np
import tqdm

try:
    from irods.session import iRODSSession
    from irods.exception import DataObjectDoesNotExist
    from irods.keywords import FORCE_FLAG_KW
    have_irods = True
except ImportError:
    have_irods = False


# to keep track of printing download progress
last_percent_printed = None


class DownloadProgressBar(tqdm.tqdm):
    """
    A progress bar for downloads using tqdm
    based on https://stackoverflow.com/questions/15644964/python-progress-bar-and-downloads
    """

    def __init__(self, *args, **kwargs):
        """
        """
        # this __init__ is only here to avoid the docstring of tqdm.tdqm's __init__ showing up
        # in the docs of arts_tools
        super(DownloadProgressBar, self).__init__(*args, **kwargs)

    def update_to(self, nbyte=1, blocksize=1, totalsize=None):
        if totalsize is not None:
            self.total = totalsize
        self.update(nbyte * blocksize - self.n)


# this function is no longer used but may be useful in the future
def format_bytes(nbytes):
    """
    Format a number in bytes with a prefix

    :param int nbytes: number of bytes (>0)
    :return: number of MB/GB etc (float), unit (MB/GB etc) (str)
    """
    # prefixes up to exabyte
    prefix_letters = ' kMGTPE'

    if nbytes < 1:
        # special case, because log10(nbytes) < 0 for nbytes < 1
        prefix_tier = 0
    else:
        prefix_tier = int(np.log10(nbytes) / 3)  # 3 = log10(1000)

    try:
        letter = prefix_letters[prefix_tier]
        scaling = 1000 ** prefix_tier
    except IndexError:
        # unlikely to run into files over 1000 exabytes, but we wouldn't want an unexplained
        # IndexError when someone puts in a very large number
        letter = prefix_letters[-1]
        scaling = 1000 ** (len(prefix_letters) - 1)
    # do the scaling
    return nbytes / scaling, letter + 'B'


# this function is no longer used but may be useful in the future
def print_progress(nblock, nbyte_per_block, nbyte_total, step=5):
    """
    Print download progress of urllib.request.urlretrieve command

    :param nblock:
    :param nbyte_per_block:
    :param total_bytes:
    :param int step: progress is only printed if the percentage is a multiple of step
    """
    global last_percent_printed
    nbyte_done = nblock * nbyte_per_block
    percent_done = int(100 * nbyte_done / nbyte_total)

    # print only if not done already and multiple of step
    if last_percent_printed != percent_done and int(percent_done % step) == 0:
        # get scaled value prefix to use for output (kB, MB etc)
        done_scaled, done_unit = format_bytes(nbyte_done)
        total_scaled, total_unit = format_bytes(nbyte_total)

        print(f"{done_scaled:.0f} {done_unit} / {total_scaled:.0f} {total_unit} ({percent_done:.0f}%)")
        last_percent_printed = percent_done
    return


def download_url(url, output_folder=None, overwrite=False, verbose=False):
    """
    Download file from given url

    :param url: URL to download file from
    :param str output_folder: Output folder (Default: current directory)
    :param bool overwrite: Overwrite output file if it already exists
    :param bool verbose: Print download progress
    """
    # create output folder and get full path to output file
    if output_folder is None:
        output_folder = os.getcwd()
    os.makedirs(output_folder, exist_ok=True)
    output_file = os.path.join(output_folder, os.path.basename(url))

    # check if the file already exists
    if os.path.isfile(output_file) and not overwrite:
        print(f"File already exists: {output_file}")
        return

    # download the file
    try:
        # with or without progress bar
        if verbose:
            with DownloadProgressBar(unit='B', unit_scale=True, miniters=1, desc=os.path.basename(url)) as progress_bar:
                urllib.request.urlretrieve(url, filename=output_file, reporthook=progress_bar.update_to)
        else:
            urllib.request.urlretrieve(url, filename=output_file)
    except urllib.error.HTTPError as e:
        # add filename to error message
        e.msg = f"URL not found: {url}"
        raise
    return


def download_irods(path, output_folder=None, overwrite=False):
    """
    Download file from ALTA using iRODS

    :param path: Path to file on iRODS server
    :param str output_folder: Output folder (Default: current directory)
    :param bool overwrite: Overwrite output file if it already exists
    """
    # check if iRODS tools are available
    if not have_irods:
        print("iRODS tools not available; reinstall arts_tools "
              "with pip install --upgrade arts_tools[irods]")
        return

    # get the irods environment file
    try:
        env_file = os.environ['IRODS_ENVIRONMENT_FILE']
    except KeyError:
        env_file = os.path.expanduser('~/.irods/irods_environment.json')
    if not os.path.isfile(env_file):
        print("iRODS environment file not found. Set IRODS_ENVIRONMENT_FILE if the file is not in the default "
              "location. Also make sure your iRODS environment is set up properly (with iinit)")
        return

    # create output folder and get full path to output file
    if output_folder is None:
        output_folder = os.getcwd()
    os.makedirs(output_folder, exist_ok=True)
    output_file = os.path.join(output_folder, os.path.basename(path))

    # check if the file already exists
    if os.path.isfile(output_file) and not overwrite:
        print(f"File already exists: {output_file}")
        return
    if overwrite:
        download_options = {FORCE_FLAG_KW: ''}
    else:
        download_options = {}

    # create SSL context
    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=None, capath=None, cadata=None)
    # open an iRODS session
    with iRODSSession(irods_env_file=env_file, ssl_context=ssl_context) as session:
        try:
            session.data_objects.get(path, output_folder, **download_options)
        except DataObjectDoesNotExist:
            # add filename to error message
            raise DataObjectDoesNotExist(f"File not found: {path}")



#!/usr/bin/env python3

import sys

import argparse
import warnings

import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from psrqpy import QueryATNF

from arts_tools import tools
from arts_tools.constants import CB_HPBW, REF_FREQ, NCB


def get_half_power_width(freq):
    """
    Calculate compound beam half-power width, assuming linear scaling with frequency

    :param Quantity freq: Frequency
    :return: half-power width (Quantity)
    """
    scaling = freq / REF_FREQ
    hpbw = CB_HPBW * scaling
    return hpbw.to(u.arcmin)


def main():
    # default parameters to query
    params = ["NAME", "BNAME", "RAJ", "DECJ", "DM", "P0", "S1400"]

    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="For a given pointing, determine which pulsars are in the field and "
                                                 "in which compound beam they are.")

    parser.add_argument("--ra", required=True, help="Right Ascension of pointing in hh:mm:ss.s format")
    parser.add_argument("--dec", required=True, help="Declination of pointing in dd:mm:ss.s format")
    parser.add_argument("--freq", type=float, default=1370,
                        help="Observing frequency (MHz), used to determine size of compound beams. "
                             "Ignored if max_dist is used."
                             "(Default: %(default)s)")
    parser.add_argument("--max_dist", type=float, help="Maximum distance from CB center (arcmin) to be considered "
                                                       "in the CB. (Default: twice half-power width)")
    parser.add_argument("--condition", type=str, help="Search condition when querying psrcat, e.g. "
                                                      "'S1400 > 1'")
    parser.add_argument("--more_info", action="store_true", help="Add pulsar parameters to output")
    parser.add_argument("--params", nargs="+", default=params,
                        help=f"Space-separated list of pulsar parameters to show "
                             f"when more_info=True. (Default: {' '.join(params)})")

    # print help if no arguments are given
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    # set max dist if not given
    if args.max_dist is None:
        args.max_dist = 2 * get_half_power_width(args.freq * u.MHz).to(u.arcmin).value

    # query ATNF for pulsars in a 4x4 degree FoV, which covers more than the entire Apertif FoV
    bound = [args.ra, args.dec, 4]
    # params to query should always include NAME, RAJ and DECJ
    query_params = args.params[:]
    for key in ["NAME", "RAJ", "DECJ"]:
        if key not in query_params:
            query_params.append(key)
    print("Querying ATNF")
    # for some reason the query gives a RuntimeWarning, suppress it
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        results = QueryATNF(params=query_params, circular_boundary=bound, condition=args.condition)

    if len(results) == 0:
        print("No pulsars found")
        sys.exit()

    # find the pointing of each CB
    pointing = SkyCoord(args.ra, args.dec, unit=(u.hourangle, u.deg))
    cb_pointings = []
    for cb in range(NCB):
        cb_ra, cb_dec = tools.cb_index_to_pointing(cb, pointing.ra, pointing.dec)
        cb_pointings.append(SkyCoord(cb_ra, cb_dec))

    # check the closest CB for each pulsar
    output = []
    pulsar_found = False
    print("Locating pulsars in Apertif Compound Beams")
    for psr in results.table:
        psr_coord = SkyCoord(psr["RAJ"], psr["DECJ"], unit=(u.hourangle, u.deg))
        # get separation from each CB and find minimum
        separations = [psr_coord.separation(beam).to(u.arcmin).value for beam in cb_pointings]
        # index of lowest separation is best CB
        best_cb = np.argmin(separations)
        # check the separation itself
        sep = separations[best_cb]
        # if too far away, skip this pulsar
        # sep must be in acrmin here
        if sep > args.max_dist:
            continue
        # a good pulsar was found
        pulsar_found = True
        # store info
        output.append([best_cb, sep, psr])

    if not pulsar_found:
        print("No pulsars found")
        sys.exit()

    # print info sorted by CB
    output = np.array(output)
    order = np.argsort(output[:, 0])

    print("Pulsars found:")
    for cb, sep, psr in output[order]:
        print(f"PSR {psr['NAME']} in CB{cb:02d}, separation from CB centre: {sep:.2f}\"")
        if args.more_info:
            for p in args.params:
                # get value and unit
                value = psr[p]
                unit = psr.columns[p].unit
                dtype = psr.columns[p].dtype
                # format according to dtype
                if dtype == np.float:
                    formatted_value = f" {p} = {value:.2f}"
                else:
                    formatted_value = f" {p} = {value}"
                # add unit
                if unit is not None:
                    formatted_value += f" {unit}"
                # add newline
                print(formatted_value)
            print()



#!/usr/bin/env python3

import os
import sys
import argparse

import logging
import numpy as np


#: fits file alignment block size in bytes
ALIGN = 2880
#: NAXIS1 in original data
NAXIS1 = 30220
#: Number of frequency channels in original data
NCHAN = 384
#: Number of time samples per subint in original data
NSAMP = 500
#: Number of seconds per subint in original data
TSUBINT = 1.024


def get_header(fname):
    """
    Read header from fits file, assuming there are 2 HDUs

    :param str fname: Path to fits file
    :return: raw header (bytes), header size in bytes (int)
    """
    # read increasinly bigger parts of the file until we encounter two END cards
    # an END card is on a line of its own
    # Each line is 80 characters, so we are looking for END plus 77 spaces
    end_card = b'END' + b' ' * 77
    hdr_size_guess = 0
    while True:
        # Load raw header as bytes
        raw_header_guess = np.fromfile(fname, dtype='B', count=hdr_size_guess).tobytes()
        try:
            ind = raw_header_guess.index(end_card)
        except ValueError:
            # not found, increase guess for header size
            # step should be smaller than align value, to avoid overshooting
            hdr_size_guess += ALIGN - 1
            continue
        # repeat for the second END card, ensuring to start search after first END card
        try:
            raw_header_guess[ind + 3:].index(end_card)
        except ValueError:
            # not found, increase guess for header size again
            hdr_size_guess += ALIGN - 1
            continue
        # Second END card found, we now have the full header
        break

    # Round size up to ALIGN value so we have the full header as written to disk
    hdr_size = int(np.ceil(hdr_size_guess / float(ALIGN)) * ALIGN)
    # Read the full header
    header = np.fromfile(fname, dtype='B', count=hdr_size).tostring().decode()
    # split into lines
    header_split = '\n'.join([header[i:i + 80] for i in range(0, hdr_size, 80)])

    logging.info("Header size: {} bytes".format(hdr_size))
    logging.debug("Raw header:\n{}".format(header_split))
    return header, hdr_size


def get_data(fname, hdr_size):
    """
    Read raw data

    :param str fname: Path to fits file
    :param int hdr_size: Header size in bytes, aligned to fits block
    :return: raw data (bytes), padding (bytes), derived naxis2 (int)
    """

    # derive total data size
    raw_data_size = os.path.getsize(fname) - hdr_size
    # this is rounded up to align value, use naxis1 to get the actual number of subints (=naxis2)
    # rounding down works as long as naxis2 is larger than align block,
    # which is always the case for ARTS data
    naxis2 = int(raw_data_size / float(NAXIS1))
    # actual data size without padding
    data_size = NAXIS1 * naxis2

    # padding is the remainder to get to raw_data_size
    padding_size = raw_data_size - data_size
    # Actual padding bytes are spaces
    padding = b' ' * padding_size

    # Load the raw data
    data = np.fromfile(fname, dtype='B', count=data_size, offset=hdr_size)

    logging.info("Data size: {}  bytes with padding of {} bytes".format(data_size, padding_size))
    logging.info("NAXIS2: {}".format(naxis2))
    logging.info("Observation duration: {}s".format(naxis2 * TSUBINT))

    return data, padding, naxis2


def fix_header(header, naxis2, force=False):
    """
    Fix the header: replace NAXIS2 = 0 by correct value and replaces
    data column bits by bytes

    :param str header: full header, including padding, as single string
    :param int naxis2: new NAXIS2 value
    :param bool force: Overwrite NAXIS2 even if original value was not zero
    :return: fixed header (bytes)
    """
    # NAXIS2 fix
    # exact key we are looking for in header
    key = "NAXIS2  ="
    try:
        key_ind = header.index(key)
    except ValueError:
        logging.error("NAXIS2 key not found in header")
        sys.exit(1)
    # from the key index keep reading header until we find the associated value
    ind = key_ind + len(key)
    char = ' '
    while char == ' ':
        ind += 1
        char = header[ind]
    # NAXIS2 is zero for a broken header, check if this is indeed the case
    if char != '0':
        if force:
            logging.warning("Original NAXIS2 is not zero; forcing application of fix")
            # keep reading until the end of the value is found, indicated by a space
            while char != ' ':
                ind += 1
                char = header[ind]
            # subtract one to have value of final character
            ind -= 1
        else:
            logging.error("Original NAXIS2 is not zero, fits file should be ok already"
                          "Re-run with --force to apply the fix anyway")
            sys.exit(1)
    # ind is now the location of the last character of the value in the header
    # extract the old header line including key and value
    old_line = header[key_ind:ind + 1]
    # find the number of padding spaces between key and value
    nspace = ind - (key_ind + len(key))
    # calculate padding for new naxis2 key
    nspace_new = nspace - len(str(naxis2)) + 1
    # construct the new line
    new_line = key + ' ' * nspace_new + str(naxis2)
    # replace in header
    header_fixed = header.replace(old_line, new_line)
    logging.debug("Old header line: {}".format(old_line))
    logging.debug("New header line: {}".format(new_line))

    # bit to bytes fix
    # assume the old value is unique enough to not appear anywhere else in the header
    old_value = '{}X'.format(NSAMP * NCHAN)
    # new value has a space at the end to make length the same as old value
    new_value = '{}B '.format(int((NCHAN * NSAMP) / 8))
    logging.debug("Replacing {} by {} in header".format(old_value, new_value))
    header_fixed = header_fixed.replace(old_value, new_value)

    return header_fixed.encode()


def fix_data(data, naxis2):
    """
    Fix the data: Swap time and frequency axis, reverse frequency ordering

    :param bytes data: raw data, without padding
    :param int naxis2: NAXIS2 value
    :return: fixed data (bytes)
    """
    # Define data columns. Could attempt to get this from the raw header, but this is always the same anyway
    dtype = np.dtype([('TSUBINT', 'f8'),
                      ('OFFS_SUB', 'f8'),
                      ('LST_SUB', 'f8'),
                      ('RA_SUB', 'f8'),
                      ('DEC_SUB', 'f8'),
                      ('GLON_SUB', 'f8'),
                      ('GLAT_SUB', 'f8'),
                      ('FD_ANG', 'f4'),
                      ('POS_ANG', 'f4'),
                      ('PAR_ANG', 'f4'),
                      ('TEL_AZ', 'f4'),
                      ('TEL_ZEN', 'f4'),
                      ('DAT_FREQ', '{}f4'.format(NCHAN)),
                      ('DAT_WTS', '{}f4'.format(NCHAN)),
                      ('DAT_OFFS', '{}f4'.format(NCHAN)),
                      ('DAT_SCL', '{}f4'.format(NCHAN)),
                      ('DATA', '{}B'.format(int((NCHAN * NSAMP) / 8)))])
    # parse the raw data
    subintdata = np.frombuffer(data, dtype=dtype)
    # sanity check: number of rows should be NAXIS2
    if len(subintdata) != naxis2:
        logging.error("Unexpected mismatch: NAXIS2 ({}) and number of subints ({}) should be equal".format(
                      naxis2, len(subintdata)))
        sys.exit(1)

    # flip the frequency ordering in columns that have a value of each frequency
    for key in ('DAT_FREQ', 'DAT_SCL', 'DAT_OFFS', 'DAT_WTS'):
        subintdata[key] = subintdata[key][:, ::-1]

    # unpack the data, FITS uses big-endian
    data_unpacked = np.unpackbits(subintdata['DATA'], bitorder='big')
    # transpose to time-frequency order, original axes are (subint, chan, time)
    # then flip the frequency axis
    data_unpacked = np.swapaxes(data_unpacked.reshape((naxis2, NCHAN, NSAMP)), 1, 2)[:, :, ::-1]
    # pack into bytes. This flattens the array, restore the subint axis
    data_packed = np.packbits(data_unpacked, bitorder='big').reshape((naxis2, -1))
    # replace data column of subintdata
    subintdata['DATA'] = data_packed
    # return as byte array
    return subintdata.tobytes()


def write_file(fname, *args):
    """
    Write output file

    :param fname: path to output file
    :param args: byte array to write to file, usually header, data, and padding
    """
    try:
        with open(fname, 'wb') as f:
            for arg in args:
                f.write(arg)
    except Exception as e:
        logging.error("Failed to write output file {}: {}".format(fname, e))
        sys.exit(1)
    logging.info("Wrote output to {}".format(fname))


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="Repair ARTS 1-bit FITS files. "
                                     "These fixes are applied:\n"
                                     "1. The NAXIS2 value in the header is changed from zero to the correct value\n"
                                     "2. The data size is expressed in bytes instead of bits\n"
                                     "3. The frequency and time axes of the data are swapped\n"
                                     "4. The frequency order of the data and weights, scales, offsets, "
                                     "and frequencies columns is flipped"
                                     )
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--output', help="Output file "
                                        "(Default: input.fits -> input_fixed.fits)")
    group.add_argument('--inplace', action='store_true', help="Fix input FITS file in-place")

    parser.add_argument('--force', action='store_true', help="Apply fix even if FITS file seems good")
    parser.add_argument('--verbose', '-v', action='store_true', help="Increase verbosity")
    parser.add_argument('file', help="Path to input FITS file")

    # print help if no arguments are given
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    # set log verbosity
    if args.verbose:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO
    logging.basicConfig(format='%(asctime)s: %(message)s', level=logging_level, stream=sys.stderr)

    # set output file name
    if args.inplace:
        args.output = args.file
    elif args.output is None:
        args.output = args.file.replace('.fits', '_fixed.fits')

    # Verify fits file exists
    if not os.path.isfile(args.file):
        logging.error("File does not exist: {}".format(args.file))
        sys.exit(1)

    logging.info("Reading file {}".format(args.file))
    # read the header
    header, hdr_size = get_header(args.file)
    # read the data
    raw_data, padding, naxis2 = get_data(args.file, hdr_size)
    # fix the header
    header_fixed = fix_header(header, naxis2, args.force)
    # fix the data
    data_fixed = fix_data(raw_data, naxis2)
    # write the output file
    write_file(args.output, header_fixed, data_fixed, padding)
